#!/usr/bin/env python3
#
#  __init__.py
"""
A μ-library for constructing cascasing style sheets from Python dictionaries.
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from io import TextIOBase
from typing import IO, Any, Dict, Iterable, List, Literal, Mapping, MutableMapping, Optional, Sequence, Union

# 3rd party
import tinycss2  # type: ignore[import-untyped]
import tinycss2.ast  # type: ignore[import-untyped]
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from domdf_python_tools.words import TAB

# this package
from dict2css.helpers import em, px, rem  # noqa: F401
from dict2css.serializer import CSSSerializer

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020-2026 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "0.6.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = [
		"IMPORTANT",
		"Style",
		"dumps",
		"dump",
		"loads",
		"load",
		]

# TODO: allow int indent like json.dumps etc.

IMPORTANT = "important"
"""
The string ``'important'``.

.. latex:vspace:: 10px
"""

# Property = Union[Tuple[Union[str, int, None], str], str, int, None]
Property = Union[Sequence, str, int, float, None]

Style = Mapping[str, Property]
"""
Type annotation representing a style for :func:`~.dumps` and :func:`~.dump`.

The keys are CSS properties.

The values can be either:

* A :class:`str`, :class:`float` or :py:obj:`None`, giving the value of the property.
* A :class:`tuple` of the property's value (as above) and the priority
  such as :data:`~.IMPORTANT` (which sets ``!important`` on the property).
"""


def dumps(
		styles: Mapping[str, Union[Style, Mapping]],
		*,
		indent: str = TAB,
		trailing_semicolon: Optional[bool] = False,
		indent_closing_brace: bool = False,
		minify: bool = False,
		sort_keys: bool = False,
		check_circular: bool = True,
		none_style: Union[Literal["none"], Literal["None"]] = "none",
		) -> str:
	r"""
	Construct a cascading style sheet from a dictionary.

	``styles`` is a mapping of CSS selector strings to styles, which map property names to their values:

	.. code-block:: python

		styles = {".wy-nav-content": {"max-width": (px(1200), IMPORTANT)}}
		print(dumps(styles))

	.. code-block:: css

		.wy-nav-content {
			max-width: 1200px !important
		}

	See the :py:obj:`~.Style` object for more information on the layout.

	The keys can also be media at-rules, with the values mappings of property names to their values:

	.. code-block:: python

		styles = {
			"@media screen and (min-width: 870px)": {
				".wy-nav-content": {"max-width": (px(1200), IMPORTANT)},
				},
			}
		print(dumps(styles))

	.. code-block:: css

		@media screen and (min-width: 870px) {
			.wy-nav-content {
				max-width: 1200px !important
			}
		}

	:param styles: A mapping of CSS selectors to styles.
	:param indent: The indent to use, such as a tab (``\t``), two spaces or four spaces.
	:param trailing_semicolon:  Whether to add a semicolon to the end of the final property.
	:param indent_closing_brace:
	:param minify: Minify the CSS. Overrides all other options.
	:param sort_keys: Sort dictionary keys alphabetically.
	:param check_circular: Check for circular references.
	:param none_style: Whether to represent :py:obj:`None` as ``None`` or ``none``.

	:return: The style sheet as a string.

	.. versionchanged:: 0.2.0  Added support for media at-rules.
	.. versionchanged:: 0.5.0  New implementation. Output may differ slightly from previous css-parser based one.
	.. versionchanged:: 0.6.0  Added ``none_style`` option.
	"""  # noqa: SXL001

	serializer = CSSSerializer(
			indent=indent,
			trailing_semicolon=trailing_semicolon,
			indent_closing_brace=indent_closing_brace,
			minify=minify,
			sort_keys=sort_keys,
			check_circular=check_circular,
			none_style=none_style,
			)

	css = serializer.encode(styles).rstrip()

	if css:
		return css + '\n'
	else:
		return ''


def dump(
		styles: Mapping[str, Union[Style, Mapping]],
		fp: Union[PathLike, IO],
		*,
		indent: str = TAB,
		trailing_semicolon: Optional[bool] = False,
		indent_closing_brace: bool = False,
		minify: bool = False,
		sort_keys: bool = False,
		check_circular: bool = True,
		none_style: Union[Literal["none"], Literal["None"]] = "none",
		) -> None:
	r"""
	Construct a style sheet from a dictionary and write it to ``fp``.

	.. code-block:: python

		styles = {".wy-nav-content": {"max-width": (px(1200), IMPORTANT)}}
		dump(styles, ...)

	.. code-block:: css

		.wy-nav-content {
			max-width: 1200px !important
		}

	See the :py:obj:`~.Style` object for more information on the layout.

	.. latex:clearpage::

	The keys can also be media at-rules, with the values mappings of property names to their values:

	.. code-block:: python

		styles = {
			"@media screen and (min-width: 870px)": {
				".wy-nav-content": {"max-width": (px(1200), IMPORTANT)},
				},
			}
		dump(styles, ...)

	.. code-block:: css

		@media screen and (min-width: 870px) {
			.wy-nav-content {
				max-width: 1200px !important
			}
		}

	:param styles: A mapping of CSS selectors to styles.
	:param fp: An open file handle, or the filename of a file to write to.
	:param indent: The indent to use, such as a tab (``\t``), two spaces or four spaces.
	:param trailing_semicolon:  Whether to add a semicolon to the end of the final property.
	:param indent_closing_brace:
	:param minify: Minify the CSS. Overrides all other options.
	:param sort_keys: Sort dictionary keys alphabetically.
	:param check_circular: Check for circular references.
	:param none_style: Whether to represent :py:obj:`None` as ``None`` or ``none``.

	.. versionchanged:: 0.2.0

		* ``fp`` now accepts :py:obj:`domdf_python_tools.typing.PathLike` objects,
		  representing the path of a file to write to.
		* Added support for media at-rules.

	.. versionchanged:: 0.5.0  New implementation. Output may differ slightly from previous css-parser based one.
	.. versionchanged:: 0.6.0  Added ``none_style`` option.
	"""  # noqa: SXL001

	css = dumps(
			styles,
			indent=indent,
			trailing_semicolon=trailing_semicolon,
			indent_closing_brace=indent_closing_brace,
			sort_keys=sort_keys,
			check_circular=check_circular,
			minify=minify,
			none_style=none_style,
			)

	if isinstance(fp, TextIOBase):
		fp.write(css)
	else:
		PathPlus(fp).write_clean(css)


def loads(styles: str) -> MutableMapping[str, MutableMapping[str, Any]]:
	r"""
	Parse a style sheet and return its dictionary representation.

	.. versionadded:: 0.2.0
	.. versionchanged:: 0.5.0  New implementation. Output may differ slightly from previous css-parser based one.

	:param styles:

	:return: The style sheet as a dictionary.

	.. latex:clearpage::
	"""

	stylesheet = tinycss2.parse_blocks_contents(styles, skip_comments=True, skip_whitespace=True)

	styles_dict: MutableMapping[str, MutableMapping[str, Any]] = {}

	def parse_style(style: List[tinycss2.ast.Node]) -> MutableMapping[str, Property]:
		style_dict: Dict[str, Property] = {}

		prop: Union[tinycss2.ast.ParseError, tinycss2.ast.Declaration]
		for prop in tinycss2.parse_declaration_list(style, skip_comments=True, skip_whitespace=True):
			if isinstance(prop, tinycss2.ast.ParseError):
				raise ValueError(prop)

			if prop.important:
				style_dict[prop.name.strip()] = (_serialize(prop.value), IMPORTANT)
			else:
				style_dict[prop.name.strip()] = _serialize(prop.value)

		return style_dict

	rule: tinycss2.ast.Node
	for rule in stylesheet:
		if isinstance(rule, tinycss2.ast.QualifiedRule):
			styles_dict[_serialize(rule.prelude)] = parse_style(rule.content)

		elif isinstance(rule, tinycss2.ast.AtRule):
			at_rule_styles = styles_dict[f"@{rule.at_keyword} {_serialize(rule.prelude)}"] = {}

			for child in tinycss2.parse_blocks_contents(rule.content, skip_comments=True, skip_whitespace=True):
				at_rule_styles[_serialize(child.prelude)] = parse_style(child.content)

		else:
			raise NotImplementedError(rule)

	return styles_dict


def _serialize(nodes: Iterable[tinycss2.ast.Node]) -> str:
	return tinycss2.serialize(nodes).strip()


def load(fp: Union[PathLike, IO]) -> MutableMapping[str, MutableMapping[str, Any]]:
	r"""
	Parse a cascading style sheet from the given file and return its dictionary representation.

	.. versionadded:: 0.2.0
	.. versionchanged:: 0.5.0  New implementation. Output may differ slightly from previous css-parser based one.

	:param fp: An open file handle, or the filename of a file to write to.

	:return: The style sheet as a dictionary.
	"""

	if isinstance(fp, TextIOBase):
		styles = fp.read()
	else:
		styles = PathPlus(fp).read_text()

	return loads(styles)



# stdlib
from ipaddress import IPv4Address
from typing import Dict, Mapping, MutableMapping, Optional, no_type_check

# 3rd party
import pytest
from coincidence.regressions import AdvancedDataRegressionFixture, AdvancedFileRegressionFixture
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.words import TAB

# this package
from dict2css import IMPORTANT, Style, dump, dumps, load, loads
from dict2css.helpers import em, px, rem


def boolean_option(name: str, id: str):  # noqa: A002,MAN002  # pylint: disable=redefined-builtin
	return pytest.mark.parametrize(
			name,
			[
					pytest.param(True, id=id),
					pytest.param(False, id=f"not {id}"),
					],
			)


def ternary_option(name: str, id: str):  # noqa: A002,MAN002  # pylint: disable=redefined-builtin
	return pytest.mark.parametrize(
			name,
			[
					pytest.param(True, id=id),
					pytest.param(False, id=f"not {id}"),
					pytest.param(None, id=f"unset {id}"),
					],
			)


@no_type_check
def test_dumps_not_mapping():
	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps([1, 2, 3])

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps((1, 2, 3))

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps("ABC")

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps(123)

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps(123.456)

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps(True)

	with pytest.raises(TypeError, match="Cannot convert .* to CSS"):
		dumps(None)


@no_type_check
@boolean_option("check_circular", "check_circular")
def test_dumps_unknown_type(check_circular: bool):
	with pytest.raises(ValueError, match="Object of type .* cannot be represented in CSS"):
		dumps({"the_key": IPv4Address("127.0.0.1")}, check_circular=check_circular)


@no_type_check
def test_dumps_bad_floats():
	with pytest.raises(ValueError, match="Out of range float values are not allowed:"):
		dumps({"the_key": float("inf")})

	with pytest.raises(ValueError, match="Out of range float values are not allowed:"):
		dumps({"the_key": float("-inf")})

	with pytest.raises(ValueError, match="Out of range float values are not allowed:"):
		dumps({"the_key": float("nan")})


@boolean_option("check_circular", "check_circular")
@boolean_option("sort_keys", "sort_keys")
@ternary_option("trailing_semicolon", "trailing_semicolon")
@boolean_option("indent_closing_brace", "indent_closing_brace")
@pytest.mark.parametrize(
		"indent",
		[
				pytest.param(TAB, id="tab"),
				pytest.param("  ", id='2'),
				pytest.param("    ", id='4'),
				pytest.param('', id='0'),
				],
		)
@pytest.mark.parametrize(
		"none_style",
		[
				pytest.param("none", id="normal_none"),
				pytest.param("None", id="cap_none"),
				],
		)
def test_dumps(
		advanced_file_regression: AdvancedFileRegressionFixture,
		trailing_semicolon: Optional[bool],
		indent_closing_brace: bool,
		indent: str,
		check_circular: bool,
		sort_keys: bool,
		tmp_pathplus: PathPlus,
		none_style: str,
		):
	stylesheet: Dict[str, Style] = {
			".wy-nav-content": {"max-width": (rem(1200), IMPORTANT), "z-index": 999},
			"li p:last-child": {
					"margin-bottom": (em(12), IMPORTANT),
					"margin-top": em(6),
					"font-size": px(14),
					"line-height": 1.5,
					"font-weight": (800, IMPORTANT),
					},
			"html": {"scroll-behavior": "smooth"},
			}

	css = dumps(
			stylesheet,
			indent=indent,
			trailing_semicolon=trailing_semicolon,
			indent_closing_brace=indent_closing_brace,
			check_circular=check_circular,
			sort_keys=sort_keys,
			none_style=none_style,  # type: ignore[arg-type]
			)
	advanced_file_regression.check(css, extension=".css")

	output_file = tmp_pathplus / "style.css"

	with output_file.open('w') as fp:
		dump(
				stylesheet,
				fp,
				indent=indent,
				trailing_semicolon=trailing_semicolon,
				indent_closing_brace=indent_closing_brace,
				check_circular=check_circular,
				sort_keys=sort_keys,
				)

	advanced_file_regression.check_file(output_file)

	dump(
			stylesheet,
			output_file,
			indent=indent,
			trailing_semicolon=trailing_semicolon,
			indent_closing_brace=indent_closing_brace,
			check_circular=check_circular,
			sort_keys=sort_keys,
			)

	advanced_file_regression.check_file(output_file)


def test_dump_minify(advanced_file_regression: AdvancedFileRegressionFixture, tmp_pathplus: PathPlus):
	stylesheet: Dict[str, Style] = {
			".wy-nav-content": {"max-width": (rem(1200), IMPORTANT), "box-shadow": None},
			"li p:last-child": {
					"margin-bottom": (em(12), IMPORTANT),
					"margin-top": em(6),
					},
			"html": {"scroll-behavior": "smooth"},
			}

	css = dumps(stylesheet, minify=True)
	advanced_file_regression.check(css, extension=".css")

	output_file = tmp_pathplus / "style.css"

	with output_file.open('w') as fp:
		dump(stylesheet, fp, minify=True)

	advanced_file_regression.check_file(output_file)

	dump(stylesheet, output_file, minify=True)
	advanced_file_regression.check_file(output_file)


def test_dumps_media(advanced_file_regression: AdvancedFileRegressionFixture, tmp_pathplus: PathPlus):
	stylesheet: Dict[str, MutableMapping] = {
			".wy-nav-content": {"max-width": (rem(1200), IMPORTANT)},
			"li p:last-child": {
					"margin-bottom": (em(12), IMPORTANT),
					"margin-top": em(6),
					},
			"html": {"scroll-behavior": "smooth"},
			"@media screen and (min-width: 870px)": {"li p:last-child": {"max-width": (rem(1200), IMPORTANT)}},
			}

	css = dumps(stylesheet, trailing_semicolon=True)
	advanced_file_regression.check(css, extension=".css")

	output_file = tmp_pathplus / "style.css"

	with output_file.open('w') as fp:
		dump(stylesheet, fp, trailing_semicolon=True)

	advanced_file_regression.check_file(output_file)

	dump(stylesheet, output_file, trailing_semicolon=True)
	advanced_file_regression.check_file(output_file)


def test_loads(advanced_data_regression: AdvancedDataRegressionFixture, tmp_pathplus: PathPlus):
	style = [
			".wy-nav-content {",
			"    max-width: 1200rem !important;",
			"    }",
			'',
			"li p:last-child {",
			"    margin-bottom: 12em !important;",
			"    margin-top: 6em;",
			"    }",
			'',
			"@media screen {",
			"    html {",
			"        scroll-behavior: smooth;",
			"        }",
			'}',
			]

	advanced_data_regression.check(loads('\n'.join(style)))

	stylesheet: Mapping[str, Mapping] = {
			".wy-nav-content": {"max-width": (rem(1200), IMPORTANT)},
			"li p:last-child": {
					"margin-bottom": (em(12), IMPORTANT),
					"margin-top": em(6),
					},
			"html": {"scroll-behavior": "smooth"},
			"@media screen and (min-width: 870px)": {"li p:last-child": {"max-width": (rem(1200), IMPORTANT)}},
			}

	assert loads(dumps(stylesheet)) == stylesheet

	style_file = tmp_pathplus / "style.css"
	dump(stylesheet, style_file)
	assert load(style_file) == stylesheet

	with style_file.open() as fp:
		assert load(fp) == stylesheet


def test_loads_bad_syntax():
	style = [
			".wy-nav-content {",
			"    max-width",
			"    }",
			]

	with pytest.raises(ValueError, match="<ParseError invalid>"):
		loads('\n'.join(style))


@no_type_check
def test_circular_references():
	css = {
			'a': {'b': 'c'},
			}
	css['d'] = css

	with pytest.raises(ValueError, match="Circular reference detected"):
		dumps(css)


def test_edge_cases():
	with pytest.raises(TypeError, match="keys must be strings, not"):
		dumps({1234: {'a': 'b'}})  # type: ignore[dict-item]

	with pytest.raises(ValueError, match="Property cannot be empty"):
		dumps({'a': {'b': ()}})

	assert dumps({}) == ''
	assert dumps({'a': {}}) == "a {}\n"
	assert dumps({'a': {'b': True}}, indent='') == "a {b: true}\n"
	assert dumps({'a': {'b': False}}, indent='') == "a {b: false}\n"

	assert dumps({'a': {'b': ('c', ('d', ))}}) == 'a {\n\tb: c d\n}\n'


def test_for_sphinx_toolbox():
	assert dumps({"p.source-link": {"margin-bottom": 0}}, trailing_semicolon=False) == """p.source-link {
	margin-bottom: 0
}
"""

	font_family = '"SFMono-Regular", Menlo, Consolas, Monaco, Liberation Mono, Lucida Console, monospace'
	css_dict = {"p.source-link": {"font-family": font_family}}

	assert dumps(
			css_dict,
			trailing_semicolon=False,
			) == """p.source-link {
	font-family: "SFMono-Regular", Menlo, Consolas, Monaco, Liberation Mono, Lucida Console, monospace
}
"""



#!/usr/bin/env python3
#
#  serializer.py
"""
Serializer for cascading style sheets.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Adapted from https://github.com/austinyu/ujson5
#  Copyright (c) 2025 Sir Austin
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from contextlib import contextmanager
from typing import Any, Callable, Dict, Iterable, Iterator, Literal, Mapping, Optional, Union

# 3rd party
from domdf_python_tools.words import TAB

__all__ = ["CSSSerializer"]

#: Python objects that can be serialized to CSS
Serializable = Union[dict, list, tuple, int, float, str, bool, None]

#: A callable that takes in an object that is not serializable and returns a serializable object
DefaultInterface = Union[
	Callable[[Any], dict],
	Callable[[Any], list],
	Callable[[Any], tuple],
	Callable[[Any], int],
	Callable[[Any], float],
	Callable[[Any], str],
	Callable[[Any], bool],
	Callable[[Any], Serializable]
	]

ESCAPE = re.compile(r'[\x00-\x1f\\\b\f\n\r\t]')
ESCAPE_DCT = {
		'\\': "\\\\",
		'\x08': "\\b",
		'\x0c': "\\f",
		'\n': "\\n",
		'\r': "\\r",
		'\t': "\\t",
		}
for i in range(0x20):
	ESCAPE_DCT.setdefault(chr(i), f"\\u{i:04x}")


class CSSSerializer:
	r"""
	Serializes a dictionary to CSS.

	This controls the formatting of the style sheet.

	:param indent: The indent to use, such as a tab (``\t``), two spaces or four spaces.
	:param trailing_semicolon:  Whether to add a semicolon to the end of the final property.
	:param indent_closing_brace:
	:param minify: Minify the CSS. Overrides all other options.
	:param sort_keys: Sort dictionary keys alphabetically.
	:param check_circular: Check for circular references.
	:param none_style: Whether to represent :py:obj:`None` as ``None`` or ``none``.

	.. versionchanged:: 0.5.0  New implementation. Output may differ slightly from previous css-parser based one.
	.. versionchanged:: 0.6.0  Added ``none_style`` option.

	.. autosummary-widths:: 1/4
	"""  # noqa: SXL001

	def __init__(
			self,
			*,
			indent: str = TAB,
			trailing_semicolon: Optional[bool] = None,
			indent_closing_brace: bool = False,
			minify: bool = False,
			sort_keys: bool = False,
			check_circular: bool = True,  # default: Optional[DefaultInterface] = None,
			none_style: Union[Literal["none"], Literal["None"]] = "none",
			) -> None:
		self._set_options(
				indent=indent,
				trailing_semicolon=trailing_semicolon,
				indent_closing_brace=indent_closing_brace,
				minify=minify,
				sort_keys=sort_keys,
				none_style=none_style,
				)

		if check_circular:
			self._markers: Optional[Dict[int, Any]] = {}
		else:
			self._markers = None

		# TODO self._default: Optional[DefaultInterface] = default
		self._default: Optional[DefaultInterface] = None

	def _set_options(
			self,
			indent: str = TAB,
			trailing_semicolon: Optional[bool] = None,
			indent_closing_brace: bool = False,
			minify: bool = False,
			sort_keys: bool = False,
			none_style: Union[Literal["none"], Literal["None"]] = "none",
			) -> None:

		self._sort_keys: bool = sort_keys
		self._indent_str: str = (None if minify else indent) or ''
		self.none_style = none_style

		if minify:
			self._key_separator: str = ':'
		else:
			self._key_separator = ": "

		if not indent and not minify:
			self._item_separator: str = "; "
		else:
			self._item_separator = ';'

		if minify:
			self._trailing_semicolon = False
		elif trailing_semicolon is None:
			self._trailing_semicolon = bool(self._indent_str)
		else:
			self._trailing_semicolon = trailing_semicolon

		if indent_closing_brace and indent is not None:
			self._indent_closing_brace = True
		else:
			self._indent_closing_brace = False

		self._minify = minify

	# TODO: deprecate
	def reset_style(self) -> None:  # pragma: no cover
		"""
		Reset the serializer to its default style.
		"""

		self._set_options()

	def encode(self, obj: Mapping) -> str:
		"""
		Return a CSS representation of a Python dictionary.

		:param obj: The Python dictionary to be serialized

		:returns: The CSS string representation of the Python dictionary
		"""

		return ''.join(self.iterencode(obj))

	def iterencode(self, obj: Mapping) -> Iterable[str]:
		"""
		Encode the given dictionary and yield each part of the CSS string representation.

		:param obj: The Python dictionary to be serialized

		:returns: An iterable of strings representing the CSS serialization of the Python dictionary.
		"""

		if not isinstance(obj, Mapping):
			raise TypeError(f"Cannot convert {type(obj)} to CSS")

		return self._iterencode_dict(obj, indent_level=-1)

	def default(self, obj: Any) -> Serializable:
		"""
		Override this method in a subclass to implement custom serialization for objects that are not serializable by default.

		This method should return a serializable object.
		If this method is not overridden, the encoder will raise a :exc:`ValueError` when trying to encode an unsupported object.

		:param obj: The object to be serialized that is not supported by default.

		:returns: A serializable object

		:raises: ValueError: If the object cannot be serialized
		"""

		if self._default is not None:  # pragma no cover  # TODO
			return self._default(obj)

		raise ValueError(f"Object of type {obj.__class__.__name__} cannot be represented in CSS")

	def _encode_float(self, obj: float) -> str:
		if obj != obj or obj == float("inf") or obj == float("-inf"):
			raise ValueError(f"Out of range float values are not allowed: {repr(obj)}")
		else:
			return repr(obj)

	def _encode_str(self, obj: str) -> str:

		def replace_unicode(match: re.Match) -> str:  # pragma: no cover  # TODO
			return ESCAPE_DCT[match.group(0)]

		return ESCAPE.sub(replace_unicode, obj)

	def _iterencode(self, obj: Any, indent_level: int) -> Iterable[str]:
		if isinstance(obj, str):
			yield self._encode_str(obj)
		elif obj is True:
			yield "true"
		elif obj is False:
			yield "false"
		elif obj is None:
			yield self.none_style  # either none (default; preferred) or None
		elif isinstance(obj, int):
			yield repr(obj)
		elif isinstance(obj, float):
			yield self._encode_float(obj)
		elif isinstance(obj, (list, tuple)):  # pragma: no cover  # TODO (needs default exposed)
			yield from self._iterencode_list(obj, indent_level)
		elif isinstance(obj, dict):  # pragma: no cover  # TODO (needs default exposed)
			yield from self._iterencode_dict(obj, indent_level)
		else:
			if self._markers is not None:
				marker_id: Optional[int] = id(obj)
				if marker_id in self._markers:  # pragma: no cover  # TODO (super edge case)
					raise ValueError("Circular reference detected")
				assert marker_id is not None
				self._markers[marker_id] = obj
			else:
				marker_id = None

			obj_user = self.default(obj)
			yield from self._iterencode(obj_user, indent_level)  # pragma: no cover  # TODO (needs default exposed)

			if self._markers is not None and marker_id is not None:  # pragma: no cover  # TODO (needs default exposed)
				del self._markers[marker_id]

	def _iterencode_list(
			self,
			obj: Union[list, tuple, None],
			indent_level: int,
			) -> Iterable[str]:

		# this package
		from dict2css import IMPORTANT

		if not obj:
			raise ValueError("Property cannot be empty")

		if self._markers is not None:
			marker_id: Optional[int] = id(obj)
			if marker_id in self._markers:  # pragma: no cover  # TODO (super edge case)
				raise ValueError("Circular reference detected")
			assert marker_id is not None
			self._markers[marker_id] = obj
		else:
			marker_id = None

		first: bool = True
		for value in obj:
			if first:
				yield ''
				first = False
			else:
				yield ' '

			if value == IMPORTANT:
				yield "!important"
			else:
				yield from self._iterencode(value, indent_level)

		if self._markers is not None and marker_id is not None:
			del self._markers[marker_id]

	def _iterencode_dict(
			self,
			obj: Mapping[str, Any],
			indent_level: int,
			) -> Iterable[str]:
		if not obj:
			if indent_level >= 0:
				yield "{}"

			return

		if self._markers is not None:
			marker_id: Optional[int] = id(obj)
			if marker_id in self._markers:
				raise ValueError("Circular reference detected")
			assert marker_id is not None
			self._markers[marker_id] = obj
		else:
			marker_id = None

		if indent_level >= 0:
			yield '{'

		minify_indent = False
		if self._minify or not self._indent_str:
			# No indent
			newline_indent: Optional[str] = None

			if indent_level == -1:
				indent_level += 1
				minify_indent = True

		else:
			indent_level += 1
			newline_indent = '\n' + self._indent_str * indent_level
			assert newline_indent is not None

			if indent_level > 0:
				yield newline_indent

		first = True
		if self._sort_keys:
			items: Any = sorted(obj.items())
		else:
			items = obj.items()

		total_items: int = len(items)
		for idx, (key, value) in enumerate(items):
			if not isinstance(key, str):
				raise TypeError(f"keys must be strings, not {key.__class__.__name__}")

			if first:
				first = False
			elif newline_indent is not None:
				yield newline_indent  # we do not need to yield anything if indent == 0
			yield self._encode_str(key)

			if isinstance(value, dict):
				if not self._minify:
					yield ' '

				yield from self._iterencode_dict(value, indent_level)
			else:
				yield self._key_separator

				if isinstance(value, (list, tuple)):
					yield from self._iterencode_list(value, indent_level)
				else:
					yield from self._iterencode(value, indent_level)

				if idx != total_items - 1 or self._trailing_semicolon:
					yield self._item_separator

		if self._indent_str:
			indent_level -= 1
			yield '\n' + self._indent_str * indent_level
		elif minify_indent:
			indent_level -= 1

		if indent_level >= 0:
			if self._indent_closing_brace:
				yield self._indent_str

			yield '}'

			if not self._minify:
				yield '\n'

		if self._markers is not None and marker_id is not None:
			del self._markers[marker_id]

	# TODO: deprecate
	@contextmanager
	def use(self) -> Iterator:  # pragma: no cover
		"""
		No-op. Deprecated.
		"""

		yield



#!/usr/bin/env python3
#
#  helpers.py
"""
Helper functions.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Union

__all__ = ["em", "px", "rem"]


def px(val: Union[int, float, str]) -> str:
	"""
	Helper function to format a number as a value in pixels.

	:param val:
	"""

	return f"{val}px"


def em(val: Union[int, float, str]) -> str:
	"""
	Helper function to format a number as a value in em.

	:param val:
	"""

	return f"{val}em"


def rem(val: Union[int, float, str]) -> str:
	"""
	Helper function to format a number as a value in rem.

	:param val:
	"""

	return f"{val}rem"
